@@ -24,9 +24,12 @@
 #endif
 
 #ifndef WIN32
-#ifdef HAVE_SYSTEMD
+/* For now link with systemd unconditionaly (all Fedora versions are using it,
+ * Archlinux also). But if someone needs no systemd in dependencies,
+ * it can be easily turned off, check the code in main() - conditions on
+ * getenv("NOTIFY_SOCKET").
+ */
 #include <systemd/sd-daemon.h>
-#endif
 #else // !WIN32
 // parameters for a client pipe thread
 struct thread_param {
@@ -217,6 +220,9 @@ int mainloop(struct db_daemon_data *d) {
                     }
                     d->multiread_requested = 1;
                 } else {
+                    /* do not send further updates, until VM's daemon restart
+                     * and re-sync */
+                    d->remote_connected = 0;
                     break;
                 }
                 break;
@@ -442,6 +448,9 @@ int mainloop(struct db_daemon_data *d) {
                     break;
                 }
             }
+            /* trigger pending data write */
+            if (libvchan_buffer_space(d->vchan))
+                write_vchan_or_client(d, NULL, NULL, 0);
             while (libvchan_data_ready(d->vchan)) {
                 if (!handle_vchan_data(d)) {
                     fprintf(stderr, "FATAL: vchan data processing failed\n");
@@ -530,9 +539,17 @@ int init_server_socket(struct db_daemon_data *d) {
 
 int init_vchan(struct db_daemon_data *d) {
     if (d->vchan) {
+        buffer_free(d->vchan_buffer);
         libvchan_close(d->vchan);
         d->vchan = NULL;
     }
+    d->vchan_buffer = buffer_create();
+    if (!d->vchan_buffer) {
+        fprintf(stderr, "vchan buffer allocation failed\n");
+        return 0;
+    }
+    d->vchan_pending_hdr.type = QDB_INVALID_CMD;
+
     if (d->remote_name) {
         /* dom0 part: listen for connection */
         if (d->remote_domid == 0) {
@@ -657,10 +674,15 @@ static void vchan_logger(IN int logLevel, IN const CHAR *function, IN const WCHA
 
 #endif
 
+#ifndef FUZZING_BUILD_MODE_UNSAFE_FOR_PRODUCTION
 int main(int argc, char **argv) {
+#else
+int fuzz_main(int argc, char **argv) {
+#endif
     struct db_daemon_data d;
 #ifndef WIN32
     int ready_pipe[2] = {0, 0};
+    pid_t pid;
 #endif
     int ret;
 
@@ -681,11 +703,7 @@ int main(int argc, char **argv) {
      * sucessful start */
     /* FIXME: OS dependent code */
 #ifndef WIN32
-#ifdef HAVE_SYSTEMD
     if (!getenv("NOTIFY_SOCKET")) {
-#else
-    if (1) {
-#endif
         char buf[6];
         char log_path[MAX_FILE_PATH];
         int log_fd;
@@ -695,7 +713,7 @@ int main(int argc, char **argv) {
             perror("pipe");
             exit(1);
         }
-        switch (fork()) {
+        switch (pid = fork()) {
             case -1:
                 perror("fork");
                 exit(1);
@@ -763,7 +781,7 @@ int main(int argc, char **argv) {
                       NULL, // notification handler
                       NULL // notification context
                       );
-#else /* WIN32 */
+#else
     if (!init_vchan(&d)) {
         fprintf(stderr, "FATAL: vchan initialization failed\n");
         exit(1);
@@ -787,12 +805,9 @@ int main(int argc, char **argv) {
 
     /* now ready for serving requests, notify parent */
     /* FIXME: OS dependent code */
-#ifdef HAVE_SYSTEMD
     if (getenv("NOTIFY_SOCKET")) {
         sd_notify(1, "READY=1");
-    } else
-#endif /* HAVE_SYSTEMD */
-    {
+    } else {
         if (write(ready_pipe[1], "ready", strlen("ready")) != strlen("ready"))
             perror("failed to notify parent");
         close(ready_pipe[1]);
@@ -801,7 +816,7 @@ int main(int argc, char **argv) {
     create_pidfile(&d);
 
     ret = !mainloop(&d);
-#endif /* !WIN32 */
+#endif
 
     if (d.vchan)
         libvchan_close(d.vchan);
